<?php

/**
 * WC_Gateway_Bankful Plugin Class
 *
 * @package Bankful with WooCommerce Payment on hosted checkout.
 * @since   2.0.0
 */

/**
 * The class itself, please note that it is inside plugins_loaded action hook
 */
class WC_Gateway_Bankful extends WC_Payment_Gateway {

	/**
	 * API endpoint
	 *
	 * @var string
	 */
	public $version = '2.0.6';

	/**
	 * API endpoint
	 *
	 * @var string
	 */
	public $api_endpoint;

	/**
	 * API endpoint for recurring payment
	 *
	 * @var string
	 */
	public $api_endpoint_subscription;

	/**
	 * API endpoint fot REFUND Transaction
	 *
	 * @var string
	 */
	public $api_endpoint_refund;

	/**
	 * Whether or not logging is enabled
	 *
	 * @var bool
	 */
	public static $log_enabled = false;

	/**
	 * Logger instance
	 *
	 * @var WC_Logger
	 */
	public static $log = false;

	/**
	 * Whether the debug mode is enabled.
	 *
	 * @var bool
	 */
	public $debug;

	/**
	 * API username for merchant
	 *
	 * @var string
	 */
	public $req_username;

	/**
	 * API password for merchant
	 *
	 * @var string
	 */
	public $req_password;

	/**
	 * Whether or not Test server is enabled
	 *
	 * @var bool
	 */
	public $testEnv;

	/**
	 * Return Url to update order status on WooCommerce
	 *
	 * @var bool
	 */
	public $notify_url;

	/**
	 * Whether the Order is Subscription order or not.
	 *
	 * @var bool
	 */
	public $is_subscription = false;

	/**
	 * Class constructor
	 */
	public function __construct() {
		global $WP_SITE_URL;
		$this->id = 'bankful_hosted_gateway'; // payment gateway plugin ID
		
		/**
		 * Filter hook to modify Bankful Icon
		 * 
		 * @since version 2.0.0
		 */
		$this->icon = apply_filters('woocommerce_bankful_hosted_icon', plugins_url() . '/bankful/assets/img/bankful_logo_whitebg.svg');
		$this->has_fields = false;
		$this->method_title = 'Bankful Payment Gateway';
		$this->order_button_text = __('Proceed to Bankful', 'woocommerce');
		$this->method_description = 'Pay securely with Bankful Payment Gateway.'; // will be displayed on the options page

		// gateway supports subscriptions, saved payment methods,
		$this->supports = array(
			'products',
			'subscriptions',
			'subscription_cancellation',
			'subscription_suspension',
			'subscription_reactivation',
			'subscription_amount_changes',
			'subscription_date_changes',
			'multiple_subscriptions',
			'refunds',
		);

		// Method with all the options fields
		$this->init_form_fields();

		// Load the settings.
		$this->init_settings();

		$this->title             = $this->get_option('title');
		$this->description       = $this->get_option('description');
		$this->enabled           = $this->get_option('enabled');

		$bankful_options = get_option('bankful_options');
		
		$this->testEnv = ( strtolower($bankful_options['bankful_field_environment'])==strtolower('Test') );
		
		$this->debug          = 'yes' === $this->get_option('debug', 'no');
		$this->req_username   = $this->testEnv ? ( isset( $bankful_options[ 'bankful_field_test_username' ] ) ? $bankful_options[ 'bankful_field_test_username' ] : 'sandbox-test@bankful.com' ) : ( isset( $bankful_options[ 'bankful_field_live_username' ] ) ? $bankful_options[ 'bankful_field_live_username' ] : '' );
		$this->req_password   = $this->testEnv ? ( isset( $bankful_options[ 'bankful_field_test_password' ] ) ? $bankful_options[ 'bankful_field_test_password' ] : 'tuMcC7m333J4VMsb' ) : ( isset( $bankful_options[ 'bankful_field_live_password' ] ) ? $bankful_options[ 'bankful_field_live_password' ] : '' );
		$this->notify_url     = $WP_SITE_URL . '/wp-json/bankful/v1/order-updated';

		$this->api_endpoint   = $this->testEnv ? 'https://api-qa.bankfulportal.com/front-calls/go-in/woocommerce-hosted-page' : 'https://api.paybybankful.com/front-calls/go-in/woocommerce-hosted-page';

		$this->api_endpoint_refund = $this->testEnv ? 'https://api-qa.bankfulportal.com/api/transaction/WooCommerce/' : 'https://api.paybybankful.com/api/transaction/WooCommerce/';

		$this->api_endpoint_subscription = $this->testEnv ? 'https://api-qa.bankfulportal.com/api/transaction/WooCommerce/' : 'https://api.paybybankful.com/api/transaction/WooCommerce/';

		self::$log_enabled    = $this->debug;

		// This action hook saves the settings
		add_action('woocommerce_update_options_payment_gateways_' . $this->id, array( $this, 'process_admin_options' ));

		// This action hook shows the receipt of order
		add_action('woocommerce_receipt_' . $this->id, array( $this, 'receipt_page' ));

		// This action hook shows Thankyou order page
		add_filter('woocommerce_thankyou_order_received_text', array( $this, 'order_received_text' ), 10, 2);

		add_action('woocommerce_scheduled_subscription_payment_' . $this->id, array( $this, 'scheduled_subscription_payment' ), 10, 2);

		// Adding Stylesheet
		wp_enqueue_style('bf-style', plugins_url() . '/bankful/assets/css/bf-style.css', array(), $this->version);
	}

	/**
	 * Return whether or not this gateway still requires setup to function.
	 *
	 * When this gateway is toggled on via AJAX, if this returns true a
	 * redirect will occur to the settings page instead.
	 *
	 * @since 2.0.0
	 * @return bool
	 */
	public function needs_setup() {
		return !is_email($this->req_username);
	}


	/**
	 * Logging method.
	 *
	 * @param string $message Log message.
	 * @param string $level Optional. Default 'info'. Possible values:
	 *                      emergency|alert|critical|error|warning|notice|info|debug.
	 */
	public static function log( $message, $level = 'info' ) {
		if (self::$log_enabled) {
			if (empty(self::$log)) {
				self::$log = wc_get_logger();
			}
			self::$log->log($level, $message, array( 'source' => 'bankful_gateway' ));
		}
	}

	/**
	 * Order received message.
	 *
	 * @param string $text Text to show on THank you page.
	 * @param object $order Order to process
	 * 
	 * @return string Thankyou message
	 */
	public function order_received_text( $text, $order ) {
		if ($order && $this->id === $order->get_payment_method()) {
			return esc_html__('Thank you for your payment. Your transaction has been completed, and a receipt for your purchase has been emailed to you.', 'woocommerce');
		}

		return $text;
	}

	/**
	 * Processes and saves options.
	 * If there is an error thrown, will continue to save and validate fields, but will leave the erroring field out.
	 *
	 * @return bool was anything saved?
	 */
	public function process_admin_options() {
		$saved = parent::process_admin_options();

		// Maybe clear logs.
		if ('yes' !== $this->get_option('debug', 'no')) {
			if (empty(self::$log)) {
				self::$log = wc_get_logger();
			}
			self::$log->clear('bankful_gateway');
		}

		return $saved;
	}

	/**
	 * Plugin options
	 */
	public function init_form_fields() {
		$bankful_options = get_option( 'bankful_options' );
		$this->form_fields = array(
			'enabled' => array(
				'title' => __('Enable/Disable', 'bankful_gateway'),
				'type' => 'checkbox',
				'label' => ( ( !empty($this->get_option('enabled')) && 'yes' == $this->get_option('enabled') ) ? 'Disable Bankful Gateway (Bankful Hosted)' : 'Enable Bankful Gateway (Bankful Hosted)' ),
				'default' => 'yes',
			),
			'title' => array(
				'title' => __('Title', 'bankful_gateway'),
				'type' => 'text',
				'description' => __('This controls the title which the user sees during checkout.', 'bankful_gateway'),
				'default' => __('Bankful', 'bankful_gateway'),
				'desc_tip'      => true,
			),
			'description' => array(
				'title' => __('Description', 'bankful_gateway'),
				'type' => 'textarea',
				'description' => __('This controls the description which the user sees during checkout.', 'bankful_gateway'),
				'default' => __('Pay with Bankful Hosted Payment Gateway', 'bankful_gateway'),
			),
			'debug' => array(
				'title'       => __('Debug Log', 'bankful_gateway'),
				'type'        => 'checkbox',
				'label'       => __('Enable logging', 'bankful_gateway'),
				'default'     => 'no',
				'description' => sprintf(__('Log Bankful Payment Gateway Events', 'bankful_gateway'), wc_get_log_file_path('bankful_gateway')),
			),
		);
	}

	/**
	 * Get Bankful Payment Gateway Args
	 *
	 * @param object $order
	 * 
	 * @return array
	 */
	public function get_bf_checkout_args( $order ) {
		global $WP_SITE_URL;
		global $woocommerce;
		global $wp_version;
		global $BF_PLUGIN_SOURCE;

		$order_id = $order->get_id();

		if ('yes' == $this->debug) {
			$this->log('Generating payment form for order ' . $order->get_order_number());
		}

		$is_subscription = false;
		if (function_exists('wcs_order_contains_subscription') && wcs_order_contains_subscription($order_id)) {
			$is_subscription = true;
		} elseif (function_exists('wcs_order_contains_renewal') && wcs_order_contains_renewal($order_id)) {
			$is_subscription = true;
		} else {
			$is_subscription = false;
		}

		$signature = md5($this->req_password);
		$bf_checkout_args = array();

		$bf_checkout_args['req_username']       = $this->req_username;
		$bf_checkout_args['signature']          = $signature;
		$bf_checkout_args['cartName']           = 'WooCommerce-Hosted-Page';
		$bf_checkout_args['authenticity_token'] = 'kcDyh8jCxl6IL3SCDVBtCjTZjMbDQ/7uT9tdit2hsUiPHiFkcC4lMJsUt6BX2OeKYcuJYTR+ENYmyjWpdsEe8g==';
		$bf_checkout_args['order_id']           = $order_id;
		$bf_checkout_args['request_amount']     = $order->get_total();
		$bf_checkout_args['x_url_cancel']       = $order->get_cancel_order_url();
		$bf_checkout_args['callback_url']       = $this->notify_url;
		$bf_checkout_args['redirect_success']   = $WP_SITE_URL . '/wp-json/bankful/v1/redirect-pending/';
		$bf_checkout_args['x_url_failed']       = $WP_SITE_URL . '/wp-json/bankful/v1/redirect-pending/';

		// $bf_checkout_args['x_url_complete']     = add_query_arg('key', $order->get_order_key(), $order->get_checkout_order_received_url(true));
		$bf_checkout_args['x_url_complete']     = $this->get_return_url( $order );
		$bf_checkout_args['request_currency']   = get_woocommerce_currency();
		$bf_checkout_args['cust_fname']         = sanitize_text_field($order->get_billing_first_name());
		$bf_checkout_args['cust_lname']         = sanitize_text_field($order->get_billing_last_name());
		$bf_checkout_args['bill_addr']          = sanitize_text_field($order->get_billing_address_1());
		$bf_checkout_args['bill_addr_2']        = sanitize_text_field($order->get_billing_address_2());
		$bf_checkout_args['bill_addr_city']     = sanitize_text_field($order->get_billing_city());
		$bf_checkout_args['bill_addr_state']    = sanitize_text_field($order->get_billing_state());
		$bf_checkout_args['bill_addr_country']  = sanitize_text_field($order->get_billing_country());
		$bf_checkout_args['bill_addr_zip']      = sanitize_text_field($order->get_billing_postcode());
		$bf_checkout_args['cust_phone']         = sanitize_text_field($order->get_billing_phone());
		$bf_checkout_args['cust_email']         = sanitize_text_field($order->get_billing_email());
		$bf_checkout_args['remote_ip']          = $order->get_customer_ip_address();
		$bf_checkout_args['save_card_on_file']  = $is_subscription ? 'Y' : 'N';

		$bf_checkout_args['woo_version']        = sanitize_text_field($woocommerce->version);
		$bf_checkout_args['wp_version']         = $wp_version;
		$bf_checkout_args['php_version']        = phpversion();
		$bf_checkout_args['bf_plugin_version']  = $this->version;
		$bf_checkout_args['bf_plugin_name']     = sanitize_text_field('Bankful Payments Hosted');
		$bf_checkout_args['bf_plugin_source']   = $BF_PLUGIN_SOURCE; // bankful_direct || woo_market

		/**
		 * Filter hook to modify Bankful checkout Arguments to pass to bankful hosted page
		 * 
		 * @since version 2.0.0
		 */
		$bf_checkout_args = apply_filters('woocommerce_bankful_hosted_args', $bf_checkout_args);

		return $bf_checkout_args;
	}

	/**
	 * Generate the bf_checkout button link
	 *
	 * @param string $order_id
	 * 
	 * @return Form Hidden checkout form to pass data.
	 */
	public function generate_bf_checkout_form( $order_id ) {

		$order = new WC_Order($order_id);

		$bf_checkout_args = $this->get_bf_checkout_args($order);

		$bf_checkout_args_array = array();

		foreach ($bf_checkout_args as $key => $value) {
			$bf_checkout_args_array[] = '<input type="hidden" name="' . esc_attr($key) . '" value="' . esc_attr($value) . '" />';
		}

		wc_enqueue_js('
            jQuery("body").block({
                message: "' . esc_js(__('Thank you for your order. We are now redirecting you to Bankful Gateway to make payment.', 'bankful_hosted')) . '",
                baseZ: 99999,
                overlayCSS:
                {
                    background: "#fff",
                    opacity: 0.6
                },
                css: {
                    padding:        "20px",
                    zindex:         "9999999",
                    textAlign:      "center",
                    color:          "#555",
                    border:         "3px solid #aaa",
                    backgroundColor:"#fff",
                    cursor:         "wait",
                    lineHeight:     "24px",
                }
            });
            jQuery("#submit_bf_checkout_payment_form").click();
        ');

		return '<form action="' . $this->api_endpoint . '" method="post" id="bf_payment_form" target="_top">'
			. implode('', $bf_checkout_args_array) .
			'<input type="submit" class="button alt" id="submit_bf_checkout_payment_form" value="' . __('Pay via Bankful', 'bankful_hosted') . '" /> <a class="button cancel" href="' . esc_url($order->get_cancel_order_url()) . '">' . __('Cancel order &amp; restore cart', 'bankful_hosted') . '</a>
            </form>';
	}

	/**
	 * Process the payment and return the result.
	 *
	 * @param  int $order_id Order ID.
	 * @return array
	 */
	public function process_payment( $order_id ) {

		$order          = wc_get_order($order_id);

		return array(
			'result'   => 'success',

			/**
			 * Filter hook to modify Bankful redirect params before redirecting
			 * 
			 * @since version 2.0.0
			 */
			'redirect' => apply_filters('process_payment_redirect', $order->get_checkout_payment_url(true), $order),
		);
	}

	/**
	 * Output for the order received page.
	 *
	 * @return void
	 */
	public function receipt_page( $order ) {

		echo '<p>Thank you for your order, please click the button below to pay with Bankful.</p>';
		
		// ./vendor/bin/phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped
		// echo 
		echo wp_kses(
			$this->generate_bf_checkout_form($order),
			array(
				'form'  => array(
					'action' => true,
					'method' => true,
					'id' => true,
					'target' => true,
				),
				'a'   => array(
					'class' => true,
					'href' => true,
				),
				'input' => array(
					'type' => true,
					'name' => true,
					'value' => true,
					'id' => true,
					'class' => true,
				),
			)
		);
	}

	/**
	 * Process the payment and return the result.
	 *
	 * @param int $order_id Order ID.
	 * @param int $amount Order ID.
	 * 
	 * @return bool
	 */
	public function process_refund( $order_id, $amount = null, $reason = '' ) {
		global $woocommerce;
		global $wp_version;
		global $BF_PLUGIN_SOURCE;

		$order  = wc_get_order($order_id);
		
		// If it's something else such as a WC_Order_Refund, we don't want that.
		if (!is_a($order, 'WC_Order')) {
			return new WP_Error('wc-order', __('Provided ID is not a WC Order', 'bankful_hosted'));
		}

		// $bf_order_id = get_post_meta($order_id, '_transaction_id', true);
		$bf_order_id = $order->get_meta('_transaction_id');

		if (empty($bf_order_id)) {
			return new WP_Error('wc-order', __('Bankful Provided Transaction ID not found.', 'bankful_hosted'));
		}

		if ('refunded' == $order->get_status()) {
			return new WP_Error('wc-order', __('Order has been already refunded', 'bankful_hosted'));
		}

		$params = array(
			'transaction_type' => 'REFUND',
			'amount' => $amount,
			'request_ref_po_id' => $bf_order_id,
			'woo_version'       => $woocommerce->version,
			'wp_version'        => $wp_version,
			'php_version'       => phpversion(),
			'bf_plugin_version' => $this->version,
			'bf_plugin_name'    => 'Bankful Payments Hosted',
			'bf_plugin_source'  => $BF_PLUGIN_SOURCE,
			'req_username'      => $this->req_username,
			'req_password'      => $this->req_password,
		);

		$url = $this->api_endpoint_refund;

		$response = wp_remote_post(
			$url,
			array(
				'body' => http_build_query($params),
				'headers'     => array(
					'Content-Type' => 'application/x-www-form-urlencoded',
				),
			)
		);

		$body = wp_remote_retrieve_body($response);

		if (!is_wp_error($response)) {

			$body = json_decode($body, false);

			if (
				isset($body->TRANS_STATUS_NAME)
				&& 'APPROVED' == $body->TRANS_STATUS_NAME
			) {

				$this->log('====================== process_payment PAYMENT APPROVED for Order: ' . $order_id, 'info');

				$order->add_order_note( 'Bankful Payment refund completed. Refund Transaction ID:-' . $body->TRANS_RECORD_ID );

				// Redirect
				return true;
			} else if (!empty($body->ERROR_MESSAGE) || !empty($body->API_ADVICE)) {

				$error_message = 'Order already refunded or something went wrong';

				// Add note.
				/* translators: %s: Transaction ID */
				$order->add_order_note(sprintf(__('Refund Failed Transaction Order ID %s', 'wc_iveri'), $body->TRANS_ORDER_ID));

				// add_post_meta($order_id, '_transaction_id', $body->TRANS_ORDER_ID, true);
				$order->add_meta_data( '_transaction_id', $body->TRANS_ORDER_ID, true );
				
				return new WP_Error('wc-order', __($error_message, 'bankful_hosted'));
				// throw new Exception(__($error_message, 'bankful_hosted'));
			}
		} else {
			wc_add_notice('Connection error.', 'error');
			return new WP_Error('wc-order', __('Connection error.', 'bankful_hosted'));
		}
		return false;
	}

	/**
	 * Process the Subscription payment and return the result.
	 */
	public function process_subscription_payment( $amount_to_charge, $renewal_order ) {
		$this->log('====================== process_subscription_payment RECURRING PAYMENT', 'info');
		global $woocommerce;
		global $wp_version;

		$order = $renewal_order;

		try {

			if (function_exists('wcs_order_contains_renewal') && wcs_order_contains_renewal($order)) {
				$this->is_subscription = true;
			} else {
				throw new Exception(__('Order has no Subscription to process', 'bankful_hosted'));
			}
			$order_id = $order->get_id();
			if (function_exists('wcs_get_subscriptions_for_renewal_order')) {
				$subscriptions = wcs_get_subscriptions_for_renewal_order($order);
				$subscription = reset($subscriptions);
	
				$customer_id = $subscription->get_meta('_bankful_customer_id', true);
	
				if (empty($customer_id)) {
					$order->add_order_note('No Customer ID Found.');
					$subscription->add_order_note('No Customer Found for this transaction, Contact Bankful Support');
					$subscription->save();
					throw new Exception(__('No Customer Found for this transaction', 'bankful_hosted'));
				}
			}

			$order_param = array(
				'request_currency'  => $order->get_currency(),
				'cust_fname'        => $order->get_billing_first_name(),
				'cust_lname'        => $order->get_billing_last_name(),
				'cust_email'        => $order->get_billing_email(),
				'bill_addr'         => $order->get_billing_address_1() . ' ' . $order->get_billing_address_2(),
				'bill_addr_zip'     => $order->get_billing_postcode(),
				'bill_addr_city'    => $order->get_billing_city(),
				'bill_addr_state'   => $order->get_billing_state(),
				'bill_addr_country' => $order->get_billing_country(),
				'cart_name'         => 'WooCommerce',
				'order_id'          => $order_id,
				'xtl_order_id'      => $order_id,
				'request_action'    => 'CCAUTHCAP',
				'customer_vault_id' => $customer_id,
				'amount'            => $amount_to_charge,
				'woo_version'       => $woocommerce->version,
				'wp_version'        => $wp_version,
				'php_version'       => phpversion(),
				'bf_plugin_version' => $this->version,
				'bf_plugin_name'    => 'Bankful Payments Hosted',
				'bf_plugin_source'  => 'woo_market',
				'req_username'      => $this->req_username,
				'req_password'      => $this->req_password,
			);

			$params = array_merge($order_param);

			$url = $this->api_endpoint_subscription;

			$response = wp_remote_post(
				$url,
				array(
					'body' => http_build_query($params),
					'headers'     => array(
						'Content-Type' => 'application/x-www-form-urlencoded',
					),
				)
			);

			$body = wp_remote_retrieve_body($response);

			if (!is_wp_error($response)) {

				$body = json_decode($body, false);

				if (
					isset($body->TRANS_STATUS_NAME)
					&& 'APPROVED' == $body->TRANS_STATUS_NAME
				) {

					// Add order note
					$order->add_order_note('Subscription Payment completed with Bankful and Transaction ID:' . $body->TRANS_ORDER_ID);
					add_post_meta($order->get_id(), '_transaction_id', $body->TRANS_ORDER_ID, true);

					$body->TRANS_VALUE = floatval($body->TRANS_VALUE);

					// Payment complete add TRANS_ORDER_ID as transaction id in post_meta table
					$order->payment_complete($body->TRANS_ORDER_ID);
					wc_reduce_stock_levels($order->get_id());

					// Redirect to the thank you page
					return array(
						'result' => 'success',
						'redirect' => $this->get_return_url($order),
					);
				} else if (!empty($body->ERROR_MESSAGE) || !empty($body->API_ADVICE)) {

					$status = 'ERROR';

					// Add note
					/* translators: %s: Payment Failed Transaction ID */
					$order->add_order_note(sprintf(__('Failed Transaction ID %s', 'wc_iveri'), $body->TRANS_ORDER_ID));

					add_post_meta($order_id, '_transaction_id', $body->TRANS_ORDER_ID, true);

					// Payment failed
					/* translators: %s: Payment Failed Error */
					$order->update_status('failed', sprintf(__('Card payment failed. Payment was rejected due to an error%s', 'wc_iveri'), $order_id));

					$error_message = 'Something went wrong, please contact to your service provider.';
					if (isset($body->TRANS_STATUS_NAME) && 'DECLINED' == $body->TRANS_STATUS_NAME) {
						$error_message = ( ( !empty($body->ERROR_MESSAGE) ) ? $body->ERROR_MESSAGE : ( !empty($body->API_ADVICE) ) ) ? $body->API_ADVICE : $error_message;
					}
					// WC_Subscriptions_Manager::process_subscription_payment_failure_on_order( $order );

					throw new Exception(__($error_message, 'bankful_hosted'));
				}
			} else {
				wc_add_notice('Connection error.', 'error');
			}
		} catch (Exception $ex) { // Add log
			$this->log('ERROR Ex', 'info');
			$this->log($ex->getMessage(), 'error');

			/* translators: %s: General Failed Transaction Error */
			$order->add_order_note(sprintf(__('Error Message %s', 'wc_iveri'), $ex->getMessage()));
		}
	}

	/**
	 * Schedule Subscription for Payment
	 *
	 * @param int $order_id Order ID.
	 * 
	 */
	public function scheduled_subscription_payment( $amount_to_charge, $renewal_order ) {
		$this->process_subscription_payment($amount_to_charge, $renewal_order);
	}
}
