<?php

/**
 * WC_Bankful Gateway Plugin Class
 *
 * @package Bankful with WooCommerce Subscriptions and Simple Products Enabled
 * @since   2.0.0
 */

/**
 * The class itself, please note that it is inside plugins_loaded action hook
 */
class WC_Bankful_WS_Gateway extends WC_Payment_Gateway {

	/**
	 * API endpoint
	 *
	 * @var string
	 */
	public $version = '2.0.6';

	/**
	 * API endpoint
	 *
	 * @var string
	 */
	public $api_endpoint;

	/**
	 * API username for merchant
	 *
	 * @var string
	 */
	public $req_username;

	/**
	 * API password for merchant
	 *
	 * @var string
	 */
	public $req_password;

	/**
	 * Whether or not Test server is enabled
	 *
	 * @var bool
	 */
	public $testEnv = false;

	/**
	 * Whether or not logging is enabled
	 *
	 * @var bool
	 */
	public static $log_enabled = false;

	/**
	 * Logger instance
	 *
	 * @var WC_Logger
	 */
	public static $log = false;

	/**
	 * Whether the debug mode is enabled.
	 *
	 * @var bool
	 */
	public $debug;

	/**
	 * Whether the Order is Subscription order or not.
	 *
	 * @var bool
	 */
	public $is_subscription = false;

	/**
	 * Class constructor, more about it in Step 3
	 */
	public function __construct() {
		$this->id = 'bankful_credit_card'; // payment gateway plugin ID
		$this->icon = plugins_url() . '/bankful/assets/img/bankful_logo_whitebg.svg';
		$this->has_fields = true;
		$this->method_title = 'Bankful (Credit Card)';
		$this->method_description = 'Bankful Gateway with support of WooCommerce Subscriptions Plugin';

		$this->is_subscription = false;

		// gateway supports subscriptions, saved payment methods,
		$this->supports = array(
			'products',
			'subscriptions',
			'subscription_cancellation',
			'subscription_suspension',
			'subscription_reactivation',
			'subscription_amount_changes',
			'subscription_date_changes',
			'multiple_subscriptions',
			'refunds',
		);

		// Method with all the options fields.
		$this->init_form_fields();

		// Load the settings.
		$this->init_settings();

		$this->title = $this->get_option('title');
		$this->description = $this->get_option('description');
		$this->enabled = $this->get_option('enabled');
		
		$bankful_options = get_option('bankful_options');
		
		$this->testEnv = ( strtolower($bankful_options['bankful_field_environment'])==strtolower('Test') );

		$this->req_username = $this->testEnv ? ( isset($bankful_options['bankful_field_test_username']) ? $bankful_options['bankful_field_test_username'] : 'sandbox-test@bankful.com' ) : ( isset($bankful_options['bankful_field_live_username']) ? $bankful_options['bankful_field_live_username'] : '' );
		$this->req_password = $this->testEnv ? ( isset($bankful_options['bankful_field_test_password']) ? $bankful_options['bankful_field_test_password'] : 'tuMcC7m333J4VMsb' ) : ( isset($bankful_options['bankful_field_live_password']) ? $bankful_options['bankful_field_live_password'] : '' );

		$this->api_endpoint = $this->testEnv ? 'https://api-qa.bankfulportal.com/api/transaction/WooCommerce/' : 'https://api.paybybankful.com/api/transaction/WooCommerce/';

		$this->debug        = 'yes' === $this->get_option('debug', 'no');

		self::$log_enabled  = $this->debug;

		// This action hook saves the settings
		add_action('woocommerce_update_options_payment_gateways_' . $this->id, array( $this, 'process_admin_options' ));
		add_action('woocommerce_scheduled_subscription_payment_' . $this->id, array( $this, 'scheduled_subscription_payment' ), 10, 2);
	}

	/**
	 * Logging method.
	 *
	 * @param string $message Log message.
	 * @param string $level   Optional. Default 'info'. Possible values:
	 *                        emergency|alert|critical|error|warning|notice|info|debug.
	 */
	public static function log( $message, $level = 'info' ) {
		if (self::$log_enabled) {
			if (empty(self::$log)) {
				self::$log = wc_get_logger();
			}
			self::$log->log($level, $message, array( 'source' => 'bankful_credit_card' ));
		}
	}

	/**
	 * Processes and saves options.
	 * If there is an error thrown, will continue to save and validate fields, but will leave the erroring field out.
	 *
	 * @return bool was anything saved?
	 */
	public function process_admin_options() {
		$saved = parent::process_admin_options();

		// Maybe clear logs.
		if ('yes' !== $this->get_option('debug', 'no')) {
			if (empty(self::$log)) {
				self::$log = wc_get_logger();
			}
			self::$log->clear('bankful_credit_card');
		}

		return $saved;
	}

	/**
	 * Plugin options in admin panel
	 * 
	 * @return Form Fields in Gateway Setting Screen
	 */
	public function init_form_fields() {
		// $bankful_options = get_option('bankful_options');
		$this->form_fields = array(
			'enabled' => array(
				'title'       => 'Enable/Disable',
				'label'       => ( ( !empty($this->get_option('enabled')) && 'yes' == $this->get_option('enabled') ) ? 'Disable Bankful Gateway (Credit Card)' : 'Enable Bankful Gateway (Credit Card)' ),
				'type'        => 'checkbox',
				'description' => '',
				'default'     => 'no',
			),
			'title' => array(
				'title'       => 'Title',
				'type'        => 'text',
				'description' => 'This controls the title which the user sees during checkout.',
				'default'     => 'Bankful (Credit Card)',
				'desc_tip'    => true,
			),
			'description' => array(
				'title'       => 'Description',
				'type'        => 'textarea',
				'description' => 'This controls the description which the user sees during checkout.',
				'default'     => 'Pay with your credit card via Bankful payment gateway.',
			),      
			'debug' => array(
				'title'       => 'Enable Debug Mode',
				'label'       => 'Debugging',
				'type'        => 'checkbox',
				'description' => 'Place the payment gateway in debug mode and write logs.',
				'default'     => 'no',
				'desc_tip'    => true,
			),
			// 'testEnv' => array(
			// 'title' => 'Test Environment',
			// 'label'       => 'Enable Test Mode',
			// 'type'        => 'checkbox',
			// 'description' => 'Place the payment gateway in test mode using test API keys.',
			// 'default'     => 'no',
			// 'desc_tip'    => true,
			// )
		);
		return $this->form_fields;
	}

	/**
	 * Payment Fields to show at Checkout Page
	 * 
	 */
	public function payment_fields() {
		// Display some description before the payment form
		if ($this->description) {
			if ($this->testEnv) {
				$this->description .= 'TEST ENVIRONMENT ENABLED. In test mode, you can use Test card numbers.';
				$this->description  = trim($this->description);
			}
			// display the description with <p> tags etc.
			echo wp_kses_post($this->description);
		}

		echo '<fieldset id="wc-' . esc_attr($this->id) . '-cc-form" class="wc-credit-card-form wc-payment-form" style="background:transparent;">';

		/**
		 * To add something before Bankful Credit card form.
		 *
		 * @since bankful version 2.0.0
		 */
		do_action('woocommerce_credit_card_form_start', $this->id);

		$today = gmdate('Y');
		$start = gmdate('Y');

		$form_nonce = wp_create_nonce('bf_cc_form_nonce');

		// $dateSection = '<div class="form-row form-row-first validate-required">
		// <label for="bfws_expiration">Expiry Date</label>
		// <select id="cc-exp-month" class="txt" name="bfws_exp_month">
		// <option value="01">Jan</option>
		// <option value="02">Feb</option>
		// <option value="03">Mar</option>
		// <option value="04">Apr</option>
		// <option value="05">May</option>
		// <option value="06">Jun</option>
		// <option value="07">Jul</option>
		// <option value="08">Aug</option>
		// <option value="09">Sep</option>
		// <option value="10">Oct</option>
		// <option value="11">Nov</option>
		// <option value="12">Dec</option>
		// </select>
		// <select id="cc-exp-year" class="txt" name="bfws_exp_year">';
		// for ($start; $start <= $today + 20; $start++) {
		//  $dateSection .= "<option value='" . $start . "'>$start</option>";
		// }

		// $dateSection .= '</select>
		// </div>';

		echo '<div class="fb_full_row form-row form-row-wide">
					<label class="bf_cc_labels">Card Number <span class="required">*</span></label>
					<input class="cc_input_full" id="bfws_ccNumber" name="bfws_ccNumber" type="text" autocomplete="off" pattern="[0-9]{12,19}" placeholder="Please enter valid Card Number">
				</div>
				<div class="fb_half_row form-row form-row-first validate-required">
					<label class="bf_cc_labels" for="bfws_expiration">Expiry Date <span class="required">*</span></label>
					<select id="cc-exp-month" class="txt" name="bfws_exp_month" class="cc_input_half">
						<option value="01">Jan</option>
						<option value="02">Feb</option>
						<option value="03">Mar</option>
						<option value="04">Apr</option>
						<option value="05">May</option>
						<option value="06">Jun</option>
						<option value="07">Jul</option>
						<option value="08">Aug</option>
						<option value="09">Sep</option>
						<option value="10">Oct</option>
						<option value="11">Nov</option>
						<option value="12">Dec</option>
					</select>
					<select id="cc-exp-year" class="txt" name="bfws_exp_year" class="cc_input_half">
						<option value="2024">2024</option>
						<option value="2025">2025</option>
						<option value="2026">2026</option>
						<option value="2027">2027</option>
						<option value="2028">2028</option>
						<option value="2029">2029</option>
						<option value="2030">2030</option>
						<option value="2031">2031</option>
						<option value="2032">2032</option>
						<option value="2033">2033</option>
						<option value="2034">2034</option>
						<option value="2035">2035</option>
						<option value="2036">2036</option>
						<option value="2037">2037</option>
						<option value="2038">2038</option>
						<option value="2039">2039</option>
						<option value="2040">2040</option>
						<option value="2041">2041</option>
						<option value="2042">2042</option>
						<option value="2043">2043</option>
						<option value="2044">2044</option>
					</select>
				</div>
				<div class="fb_half_row form-row form-row-last">
					<label class="bf_cc_labels">Card Code (CVV) <span class="required">*</span></label>
					<input id="bfws_cvv" name="bfws_cvv" type="password" maxlength="4" size="4" pattern="[0-9]+" required autocomplete="off" placeholder="CVV" class="cc_input_full">
				</div>
				<input type="hidden" value="' . esc_attr($form_nonce) . '" name="bf_cc_nonce" >
				<div class="clear"></div>';

		/**
		 * To add something after Bankful Credit card form.
		 *
		 * @since bankful version 2.0.0
		 */
		do_action('woocommerce_credit_card_form_end', $this->id);

		echo '<div class="clear"></div></fieldset>';
	}

	/**
	 * Validate Expiry Date entered in Checkout Card Fields
	 * 
	 * @param string||date $card_date date entered
	 * 
	 * @return boolean
	 */
	public function validate_expirydate( $card_date = null ) {
		$today = gmdate('Ym');
		$today_dt = new DateTime($today);
		$expire_dt = new DateTime($card_date);

		if ($expire_dt < $today_dt) {
			return false;
		} else {
			return true;
		}
	}

	/**
	 * Get Cart Total from Cart Object
	 * 
	 * @return float
	 */
	public function get_cart_total() {
		$final_array = array();
		$cart = WC()->cart;
		$totalAmount = $cart->total;
		$final_array['amount'] = floatval($totalAmount);
		return $final_array;
	}

	/**
	 * Get Extra Versioning Options
	 * 
	 * @return float
	 */
	public function get_versions() {
		global $woocommerce;
		global $wp_version;
		global $BF_PLUGIN_SOURCE;
		$versions_array = array();
		$versions_array['woo_version'] = $woocommerce->version;
		$versions_array['wp_version'] = $wp_version;
		$versions_array['php_version'] = phpversion();
		$versions_array['bf_plugin_version'] = $this->version;
		$versions_array['bf_plugin_name'] = 'Bankful Payments Native';
		$versions_array['bf_plugin_source'] = $BF_PLUGIN_SOURCE; // bankful_direct || woo_market
		return $versions_array;
	}

	/**
	 * Get Cart SubTotal from Cart Object
	 * 
	 * @return float
	 */
	public function get_cart_subtotal() {
		$final_array = array();
		$cart = WC()->cart;
		$subTotalAmount = $cart->subtotal;
		$final_array['subtotal_amount'] = floatval($subTotalAmount);
		return $final_array;
	}

	/**
	 * Get Cart Discount Total from Cart Object
	 * 
	 * @return float
	 */
	public function get_discount_total() {
		$final_array = array();
		$cart = WC()->cart;
		$discountAmount = $cart->get_discount_total();
		$final_array['discount_amount'] = floatval($discountAmount);
		return $final_array;
	}

	/**
	 * Get Cart Shipping Total from Cart Object
	 * 
	 * @return float
	 */
	public function get_shipping_total() {
		$final_array = array();
		$cart = WC()->cart;
		$shippingAmount = $cart->get_shipping_total();
		$final_array['shipping_amount'] = floatval($shippingAmount);
		return $final_array;
	}

	/**
	 * Get Merchant Credentials from DB
	 * 
	 * @return array
	 */
	public function merchant_credential() {
		$request_params = array(
			'req_username' => $this->req_username,
			'req_password' => $this->req_password,
		);
		$finalRequestParams = array();

		foreach ($request_params as $reqKey => $reqParamVal) {
			if (empty($request_params[$reqKey])) {
				throw new Exception(
					esc_html(
						'Merchant not configured correctly, please contact Bankful support, Thanks.',
						'bankful_credit_card'
					)
				);
			}

			$finalRequestParams[$reqKey] = trim($reqParamVal);
		}

		return $finalRequestParams;
	}

	/**
	 * Process the payment and return the result.
	 *
	 * @param int $order_id Order ID.
	 * @param int $amount Order ID.
	 * 
	 * @return bool
	 */
	public function process_refund( $order_id, $amount = null, $reason = '' ) {
		$order  = wc_get_order($order_id);
		$this->log('====================== params PAYMENT order_param Order: ' , 'info');
		
		// If it's something else such as a WC_Order_Refund, we don't want that.
		if (!is_a($order, 'WC_Order')) {
			return new WP_Error('wc-order', __('Provided ID is not a WC Order', 'bankful_credit_card'));
		}

		// $bf_order_id = get_post_meta($order_id, '_transaction_id', true);
		$bf_order_id = $order->get_meta('_transaction_id');

		if (empty($bf_order_id)) {
			return new WP_Error('wc-order', __('Bankful Provided Transaction ID not found.', 'bankful_credit_card'));
		}

		if ('refunded' == $order->get_status()) {
			return new WP_Error('wc-order', __('Order has been already refunded', 'bankful_credit_card'));
		}

		$order_param = array(
			'transaction_type' => 'REFUND',
			'amount' => $amount,
			'request_ref_po_id' => $bf_order_id,
		);

		$params = array_merge($this->merchant_credential(), $order_param, $this->get_versions());
		$this->log('====================== params PAYMENT order_param Order: ' . $order_param, 'info');
		$url = $this->api_endpoint;

		$response = wp_remote_post(
			$url,
			array(
				'body' => http_build_query($params),
				'headers'     => array(
					'Content-Type' => 'application/x-www-form-urlencoded',
				),
			)
		);

		$body = wp_remote_retrieve_body($response);

		// write_log('$body');
		// write_log($body);

		if (!is_wp_error($response)) {

			$bodyArr = $body;
			$body = json_decode($body, false);
			$bodyArr = json_decode($bodyArr, true);
			if (verifyHashSignatureFromRequest($bodyArr)) {

				if (
					isset($body->TRANS_STATUS_NAME)
					&& 'APPROVED' == $body->TRANS_STATUS_NAME
				) {

					$this->log('====================== process_payment PAYMENT APPROVED for Order: ' . $order_id, 'info');

					$order->add_order_note( 'Bankful Payment refund completed. Refund Transaction ID:-' . $body->TRANS_RECORD_ID );

					// Redirect
					return true;
				} else if (!empty($body->ERROR_MESSAGE) || !empty($body->API_ADVICE)) {

					$error_message = 'Order already refunded or something went wrong';

					// Add note.
					/* translators: %s: Transaction ID */
					$order->add_order_note(sprintf(__('Refund Failed Transaction Order ID %s', 'wc_iveri'), $body->TRANS_ORDER_ID));

					// add_post_meta($order_id, '_transaction_id', $body->TRANS_ORDER_ID, true);
					$order->add_meta_data( '_transaction_id', $body->TRANS_ORDER_ID, true );
					
					return new WP_Error('wc-order', __($error_message, 'bankful_credit_card'));
					// throw new Exception(__($error_message, 'bankful_credit_card'));
				}
			} else {
				$error_message = 'Order already refunded or something went wrong';

				// Add note.
				/* translators: %s: Transaction ID */
				$order->add_order_note(sprintf(__('Signature Mismatched Refund Failed Transaction Order ID %s', 'wc_iveri'), $body->TRANS_ORDER_ID));

				// add_post_meta($order_id, '_transaction_id', $body->TRANS_ORDER_ID, true);
				$order->add_meta_data( '_transaction_id', $body->TRANS_ORDER_ID, true );
				
				return new WP_Error('wc-order', __($error_message, 'bankful_credit_card'));
			}
		} else {
			wc_add_notice('Connection error.', 'error');
			return new WP_Error('wc-order', __('Connection error.', 'bankful_credit_card'));
		}
		return false;
	}

	/**
	 * Process the payment and return the result.
	 *
	 * @param int $order_id Order ID.
	 * 
	 */
	public function process_payment( $order_id ) {

		$this->log('====================== process_payment CAPTURE PAYMENT for Order: ' . $order_id, 'info');

		if (!isset($_POST['bf_cc_nonce'])) {
			throw new Exception(esc_html( 'Error Processing Credit Card Payment', 'bankful_credit_card'));
			die(esc_html('Security Check', 'bankful_credit_card'));
		}

		if (!wp_verify_nonce(sanitize_key($_POST['bf_cc_nonce']), 'bf_cc_form_nonce')) {
			throw new Exception(esc_html( 'Error Processing Credit Card Payment', 'bankful_credit_card'));
			die(esc_html('Security Check', 'bankful_credit_card'));
		}

		$expiry_month = isset($_POST['bfws_exp_month']) ? wc_clean($_POST['bfws_exp_month']) : null;
		$expiry_year = isset($_POST['bfws_exp_year']) ? wc_clean($_POST['bfws_exp_year']) : null;

		// we need it to get any order detailes
		$order = wc_get_order($order_id);
		global $woocommerce;
		// $order = new WC_Order( $order_id );

		$card_number = !empty(wc_clean($_POST['bfws_ccNumber']))
			? str_replace(array( ' ', '-' ), '', wc_clean($_POST['bfws_ccNumber']))
			: '';
		$card_cvc = !empty(wc_clean($_POST['bfws_cvv']))
			? wc_clean($_POST['bfws_cvv'])
			: '';

		try {
			if (empty($card_number)) {
				throw new Exception(__('Please Enter Card number', 'bankful_credit_card'));
			} elseif (empty($expiry_month) || empty($expiry_year)) {
				throw new Exception(__('Please Enter Card expiration date', 'bankful_credit_card'));
			} elseif ($this->validate_expirydate($expiry_year . $expiry_month) == false) {
				throw new Exception(__('Invalid Credit Card Expiration date', 'bankful_credit_card'));
			} elseif (empty($card_cvc)) { // check expiry date
				throw new Exception(__('Please Enter Card CVV number', 'bankful_credit_card'));
			}

			if (function_exists('wcs_get_subscriptions_for_order') && function_exists('wcs_order_contains_subscription') && wcs_order_contains_subscription($order_id)) {
				$subscriptions = wcs_get_subscriptions_for_order($order_id);
				$this->is_subscription = true;
			} elseif (function_exists('wcs_get_subscriptions_for_renewal_order') && function_exists('wcs_order_contains_renewal') && wcs_order_contains_renewal($order_id)) {
				$subscriptions = wcs_get_subscriptions_for_renewal_order($order_id);
				$this->is_subscription = true;
			} else {
				$this->is_subscription = false;
			}

			if ($this->is_subscription) {
				$subscription = reset($subscriptions);
				$this->log('====================== process_payment CAPTURE PAYMENT for Subscription: ' . $subscription->get_ID(), 'info');
			}

			$order_param = array(
				'pmt_numb' => $card_number,
				'pmt_key' => $card_cvc,
				'pmt_expiry' => $expiry_month . '/' . $expiry_year,
				'request_currency' => $order->get_currency(),
				// 'xtl_ip' => get_post_meta($order_id, '_customer_ip_address', true),
				'remote_ip' => $order->get_meta('_customer_ip_address'),
				// "xtl_ip" => "127.0.0", // For local testing purpose
				'cust_fname' => sanitize_text_field($order->get_billing_first_name()),
				'cust_lname' => sanitize_text_field($order->get_billing_last_name()),
				'cust_email' => sanitize_text_field($order->get_billing_email()),
				'bill_addr' => sanitize_text_field($order->get_billing_address_1() . ' ' . $order->get_billing_address_2()),
				'bill_addr_zip' => sanitize_text_field($order->get_billing_postcode()),
				'bill_addr_city' => sanitize_text_field($order->get_billing_city()),
				'bill_addr_state' => sanitize_text_field($order->get_billing_state()),
				'bill_addr_country' => sanitize_text_field($order->get_billing_country()),
				'ship_addr_country' => sanitize_text_field($order->get_shipping_country()),
				'ship_addr_city' => sanitize_text_field($order->get_shipping_city()),
				'ship_addr_state' => sanitize_text_field($order->get_shipping_state()),
				'ship_addr_zip' => sanitize_text_field($order->get_shipping_postcode()),
				'ship_addr' => sanitize_text_field($order->get_shipping_address_1() . ', ' . $order->get_shipping_address_2()),
				'cart_name' => 'WooCommerce',
				'order_id' => $order_id,
				'xtl_order_id' => $order_id,
				'request_action' => 'CCAUTHCAP',
				'save_card_on_file' => $this->is_subscription ? 'Y' : 'N',
			);

			$params = array_merge($this->merchant_credential(), $order_param, $this->get_versions(), $this->get_cart_total(), $this->get_cart_subtotal(), $this->get_shipping_total(), $this->get_discount_total());

			$url = $this->api_endpoint;

			$response = wp_remote_post(
				$url,
				array(
					'body' => http_build_query($params),
					'headers'     => array(
						'Content-Type' => 'application/x-www-form-urlencoded',
					),
				)
			);

			$body = wp_remote_retrieve_body($response);

			// write_logs_bf('$body 1');
			// write_logs_bf(print_r($body, true));

			if (!is_wp_error($response)) {
				$bodyArr = $body;
				$body = json_decode($body, false);
				$bodyArr = json_decode($bodyArr, true);
				if (verifyHashSignatureFromRequest($bodyArr)) {
					if (
						isset($body->TRANS_STATUS_NAME)
						&& 'APPROVED' == $body->TRANS_STATUS_NAME
					) {

						$this->log('====================== process_payment PAYMENT APPROVED for Order: ' . $order_id, 'info');

						// Add order note
						$order->add_order_note(' Payment completed with Bankful and Transaction ID:' . $body->TRANS_ORDER_ID);
						// add_post_meta($order->get_id(), '_transaction_id', $body->TRANS_ORDER_ID, true);
						$order->add_meta_data( '_transaction_id', $body->TRANS_ORDER_ID, true );

						$body->TRANS_VALUE = floatval($body->TRANS_VALUE);

						// Payment complete add TRANS_ORDER_ID as transaction id in post_meta table
						$order->payment_complete($body->TRANS_ORDER_ID);
						wc_reduce_stock_levels($order->get_id());

						if ($this->is_subscription) {
							if (isset($body->CUSTOMER_VAULT_ID) && $body->CUSTOMER_VAULT_ID) {
								$subscription->update_meta_data('_bankful_customer_id', $body->CUSTOMER_VAULT_ID);
								if (class_exists('WC_Subscriptions_Manager') && function_exists('activate_subscriptions_for_order')) {
									WC_Subscriptions_Manager::activate_subscriptions_for_order($order);
								}
								$subscription->save();
							} else {
								$order->add_order_note('Customer Vault ID not found, so next scheduled subscription transaction will not work. Contact Bankful Support for this.');
								$subscription->add_order_note('Customer Vault ID not found, so next scheduled subscription transaction will not work. Contact Bankful Support for this.');
								$subscription->save();
							}
						}

						wc_add_notice('Transaction has been completed successfully.', 'success');

						// Empty cart
						$woocommerce->cart->empty_cart();

						// Redirect to the thank you page
						return array(
							'result' => 'success',
							'redirect' => $this->get_return_url($order),
						);
					} else if (!empty($body->ERROR_MESSAGE) || !empty($body->API_ADVICE)) {

						// Add note.
						/* translators: %s: Transaction ID */
						$order->add_order_note(sprintf(__('Transaction ID %s', 'wc_iveri'), $body->TRANS_ORDER_ID));

						// add_post_meta($order_id, '_transaction_id', $body->TRANS_ORDER_ID, true);
						$order->add_meta_data( '_transaction_id', $body->TRANS_ORDER_ID, true );

						// Payment failed.
						/* translators: %s: Payment Failed Error */
						$order->update_status('failed', sprintf(__('Card payment failed. Payment was rejected due to an error%s', 'wc_iveri'), $order_id));

						$this->log('Transaction Failed', 'info');
						$this->log(print_r($body, true));

						$error_message = 'Something went wrong, please contact to your service provider.';
						if (isset($body->TRANS_STATUS_NAME) && 'DECLINED' == $body->TRANS_STATUS_NAME) {
							$error_message = ( ( !empty($body->ERROR_MESSAGE) ) ? $body->ERROR_MESSAGE : ( !empty($body->API_ADVICE) ) ) ? $body->API_ADVICE : $error_message;
						}
						wc_add_notice('Payment failed. ', 'error');

						throw new Exception(__($error_message, 'bankful_credit_card'));
					}
				} else {
					/* translators: %s: Transaction ID */
					$order->add_order_note(sprintf(__('Transaction ID %s', 'wc_iveri'), $body->TRANS_ORDER_ID));

					// add_post_meta($order_id, '_transaction_id', $body->TRANS_ORDER_ID, true);
					$order->add_meta_data( '_transaction_id', $body->TRANS_ORDER_ID, true );

					// Payment failed.
					/* translators: %s: Payment Failed Error */
					$order->update_status('failed', sprintf(__('Card payment failed. Payment was rejected due to an error%s', 'wc_iveri'), $order_id));

					$this->log('Transaction Failed', 'info');
					$this->log(print_r($body, true));

					$error_message = 'Something went wrong, please contact to your service provider.';
					if (isset($body->TRANS_STATUS_NAME) && 'DECLINED' == $body->TRANS_STATUS_NAME) {
						$error_message = ( ( !empty($body->ERROR_MESSAGE) ) ? $body->ERROR_MESSAGE : ( !empty($body->API_ADVICE) ) ) ? $body->API_ADVICE : $error_message;
					}
					wc_add_notice('Payment failed. Signature Mismatched', 'error');
					$woocommerce->cart->empty_cart();
					throw new Exception(__($error_message, 'bankful_credit_card'));
				}
			} else {
				wc_add_notice('Connection error.', 'error');
			}
		} catch (Exception $ex) { // Add log
			$this->log('ERROR Throws Exception', 'info');
			$this->log($ex->getMessage(), 'error');

			wc_add_notice($ex->getMessage(), 'error');
			return array(
				'result' => 'failed',
				'redirect' => $this->get_return_url($order),
			);
		}
		return array(
			'result' => 'failed',
			'redirect' => $this->get_return_url($order),
		);
	}

	/**
	 * Process the Subscription payment and return the result.
	 */
	public function process_subscription_payment( $amount_to_charge, $renewal_order ) {
		$this->log('====================== process_subscription_payment RECURRING PAYMENT', 'info');
		global $woocommerce;
		$order = $renewal_order;

		try {
			if (function_exists('wcs_order_contains_subscription') && function_exists('wcs_order_contains_renewal') && wcs_order_contains_renewal($order)) {
				$this->is_subscription = true;
			} else {
				throw new Exception(__('Order has no Subscription to process', 'bankful_credit_card'));
			}
			$order_id = $order->get_id();
			if (function_exists('wcs_get_subscriptions_for_renewal_order')) {
				$subscriptions = wcs_get_subscriptions_for_renewal_order($order);
				$subscription = reset($subscriptions);
	
				$customer_id = $subscription->get_meta('_bankful_customer_id', true);
	
				if (empty($customer_id)) {
					$order->add_order_note('No Customer Found : ' . $customer_id);
					$subscription->add_order_note('No Customer Found for this transaction, Contact Bankful Support');
					$subscription->save();
					throw new Exception(__('No Customer Found for this transaction', 'bankful_credit_card'));
				}
			}

			$order_param = array(
				'request_currency' => $order->get_currency(),
				'cust_fname' => $order->get_billing_first_name(),
				'cust_lname' => $order->get_billing_last_name(),
				'cust_email' => $order->get_billing_email(),
				'bill_addr' => $order->get_billing_address_1() . ' ' . $order->get_billing_address_2(),
				'bill_addr_zip' => $order->get_billing_postcode(),
				'bill_addr_city' => $order->get_billing_city(),
				'bill_addr_state' => $order->get_billing_state(),
				'bill_addr_country' => $order->get_billing_country(),
				'cart_name' => 'WooCommerce',
				'order_id' => $order_id,
				'xtl_order_id' => $order_id,
				'request_action' => 'CCAUTHCAP',
				'amount' => $amount_to_charge,
			);

			
			if (isset($customer_id)) {
				$params = array_merge($this->merchant_credential(), $order_param, $this->get_versions(), array( 'customer_vault_id'=> $customer_id ));
			} else {
				$params = array_merge($this->merchant_credential(), $order_param, $this->get_versions());
			}

			$url = $this->api_endpoint;

			$response = wp_remote_post(
				$url,
				array(
					'body' => http_build_query($params),
					'headers'     => array(
						'Content-Type' => 'application/x-www-form-urlencoded',
					),
				)
			);

			$body = wp_remote_retrieve_body($response);

			if (!is_wp_error($response)) {

				$bodyArr = $body;
				$body = json_decode($body, false);
				$bodyArr = json_decode($bodyArr, true);
				if (verifyHashSignatureFromRequest($bodyArr)) {
					if (
					isset($body->TRANS_STATUS_NAME)
					&& 'APPROVED' == $body->TRANS_STATUS_NAME
					) {

						// Add order note
						$order->add_order_note('Subscription Payment completed with Bankful and Transaction ID:' . $body->TRANS_ORDER_ID);
						// add_post_meta($order->get_id(), '_transaction_id', $body->TRANS_ORDER_ID, true);
						$order->add_meta_data( '_transaction_id', $body->TRANS_ORDER_ID, true );

						$body->TRANS_VALUE = floatval($body->TRANS_VALUE);

						// Payment complete add TRANS_ORDER_ID as transaction id in post_meta table
						$order->payment_complete($body->TRANS_ORDER_ID);
						wc_reduce_stock_levels($order->get_id());

						// Redirect to the thank you page
						return array(
							'result' => 'success',
							'redirect' => $this->get_return_url($order),
						);
					} else if (!empty($body->ERROR_MESSAGE) || !empty($body->API_ADVICE)) {

						$status = 'ERROR';

						// Add note
						/* translators: %s: Payment Failed Transaction ID */
						$order->add_order_note(sprintf(__('Failed Transaction ID %s', 'wc_iveri'), $body->TRANS_ORDER_ID));

						// add_post_meta($order_id, '_transaction_id', $body->TRANS_ORDER_ID, true);
						$order->add_meta_data( '_transaction_id', $body->TRANS_ORDER_ID, true );
						// Payment failed
						/* translators: %s: Payment Failed Error */
						$order->update_status('failed', sprintf(__('Card payment failed. Payment was rejected due to an error%s', 'wc_iveri'), $order_id));

						$error_message = 'Something went wrong, please contact to your service provider.';
						if (isset($body->TRANS_STATUS_NAME) && 'DECLINED' == $body->TRANS_STATUS_NAME) {
							$error_message = ( ( !empty($body->ERROR_MESSAGE) ) ? $body->ERROR_MESSAGE : ( !empty($body->API_ADVICE) ) ) ? $body->API_ADVICE : $error_message;
						}
						// WC_Subscriptions_Manager::process_subscription_payment_failure_on_order( $order );

						throw new Exception(__($error_message, 'bankful_credit_card'));
					}
				} else {
					// Add note
					/* translators: %s: Payment Failed Transaction ID */
					$order->add_order_note(sprintf(__('Failed Transaction ID %s', 'wc_iveri'), $body->TRANS_ORDER_ID));

					// add_post_meta($order_id, '_transaction_id', $body->TRANS_ORDER_ID, true);
					$order->add_meta_data( '_transaction_id', $body->TRANS_ORDER_ID, true );
					// Payment failed
					/* translators: %s: Payment Failed Error */
					$order->update_status('failed', sprintf(__('Signature Mismatched, Card payment failed. Payment was rejected due to altered data %s', 'wc_iveri'), $order_id));

					$error_message = 'Something went wrong, please contact to your service provider.';
					if (isset($body->TRANS_STATUS_NAME) && 'DECLINED' == $body->TRANS_STATUS_NAME) {
						$error_message = ( ( !empty($body->ERROR_MESSAGE) ) ? $body->ERROR_MESSAGE : ( !empty($body->API_ADVICE) ) ) ? $body->API_ADVICE : $error_message;
					}
					// WC_Subscriptions_Manager::process_subscription_payment_failure_on_order( $order );

					throw new Exception(__($error_message, 'bankful_credit_card'));
				}
			} else {
				wc_add_notice('Connection error.', 'error');
			}
		} catch (Exception $ex) { // Add log
			$this->log('ERROR Ex', 'info');
			$this->log($ex->getMessage(), 'error');

			/* translators: %s: General Failed Transaction Error */
			$order->add_order_note(sprintf(__('Error Message %s', 'wc_iveri'), $ex->getMessage()));
		}
	}

	/**
	 * Schedule Subscription for Payment
	 *
	 * @param int $order_id Order ID.
	 * 
	 */
	public function scheduled_subscription_payment( $amount_to_charge, $renewal_order ) {
		$this->process_subscription_payment($amount_to_charge, $renewal_order);
	}
}
